/*
 * Copyright (C) 2019 Ernesto A. Fernández <ernesto.mnd.fernandez@gmail.com>
 */

#ifndef _SPACEMAN_H
#define _SPACEMAN_H

extern u32 spaceman_size(void);
extern void set_spaceman_info(void);
extern void make_spaceman(u64 bno, u64 oid);

#endif	/* _SPACEMAN_H */
