/*
 * Copyright (C) 2021 Ernesto A. Fernández <ernesto@corellium.com>
 */

#ifndef _PARAMETERS_H
#define _PARAMETERS_H

#include <apfs/types.h>

extern u16 ip_fq_node_limit(u64 chunks);
extern u16 main_fq_node_limit(u64 blocks);

#endif /* _PARAMETERS_H */
